/***************************************************************************
 *   Copyright (C) 2004 by Alessandro Bonometti                            *
 *   bauno@bauniga.baita                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "updatedb.h"
#include "binheader.h"

void UpdateDbThread::run() {

// 	qDebug("UpdateDbThread::run(): Items in queue: %d", jobList->count() );
	
	while(!jobList->isEmpty()) {
		listLock->lock();
		job=jobList->first();
		listLock->unlock();
		if (update() ) {
		
			listLock->lock();
			jobList->remove(jobList->first());
			listLock->unlock();
			e = new UpdateDbThreadEvent(job, UpdateDbThreadEvent::UpDb_stopUpd);
			QApplication::postEvent(parent, e);
		} else {
			// Db error...
			listLock->lock();
			jobList->remove(jobList->first());
			listLock->unlock();
			e = new UpdateDbThreadEvent(job, UpdateDbThreadEvent::UpDb_Err);
			
			//set the error and exit...
			
			QApplication::postEvent(parent, e);
			break;
		}
		
// 		qDebug("Updated, now the queue has %d items", jobList->count());
	}
// 	qDebug("Update done");


	
}

bool UpdateDbThread::expire( NewsGroup * ng, int hostId) {
    Dbc *cursor;
	int lw=ng->low[hostId];
	QTime previous, current;
	previous=QTime::currentTime();
	uint progress=groupArticles;

	if ((ng->getDb()->cursor(0, &cursor, DB_WRITECURSOR))!= 0) {
		qDebug("Error creating cursor!!");
		//Set the error!
		return false;
	}
    BinHeader *bh;
    int count =0, ret;
	Dbt key, data;
	
	
	memset(&data, 0, sizeof(data));
	data.set_flags(DB_DBT_MALLOC);
	
	memset(&key, 0, sizeof(key));
	key.set_flags(DB_DBT_MALLOC);
	

	
	
	while((ret=cursor->get(&key, &data, DB_NEXT)) == 0) {
		
			
		count++;
		bh=new BinHeader((uchar*) data.get_data());
		free(data.get_data());
		
		
		switch (bh->expire(hostId, lw)) {
			case BinHeader::Delete_Unread:
				unreadArticles--;
			case BinHeader::Delete_Read:
				if ((ret=cursor->del(0))!=0)
					qDebug("Error deleting from db: %d", ret);
				else groupArticles--;
				memset(&data, 0, sizeof(data));
				data.set_flags(DB_DBT_MALLOC);
				break;
			case BinHeader::No_Delete:
				uchar *p=bh->data();
				memset(&data, 0, sizeof(data));
				data.set_data(p);
				data.set_size(bh->getRecordSize());
				
				if ((ret=cursor->put(NULL, &data, DB_CURRENT))!=0) 
					qDebug("Error updating post: %d", ret);
				// 			else qDebug("Error! %d", ret);
				delete p;
				memset(&data, 0, sizeof(data));
				data.set_flags(DB_DBT_MALLOC);
				break;
			
		}
		free(key.get_data());
		memset(&key, 0, sizeof(key));
		key.set_flags(DB_DBT_MALLOC);
		
		delete bh;
		
		if (ret != 0) {
			qDebug("Error while expiring...");
			return false;
		}
		
		current=QTime::currentTime();
		if (previous.secsTo(current) > 1) {
			e=new UpdateDbThreadEvent(job, UpdateDbThreadEvent::UpDb_Working, count, progress);
			QApplication::postEvent(parent, e);
			previous=QTime::currentTime();
			
		}
		
	} 
	
	if (ret != DB_NOTFOUND) {
		qDebug("Exited from cursor cycle with strange error %d", ret);
		return false;
	}
	

    // Begin pasted code
    if (cursor->close() != 0)
		qDebug("Error closing cursor!");

	return true;
//     qDebug("Elements in DB: %d", count);

}

bool UpdateDbThread::update()
{
	//expire newsgroup
	groupArticles=job->ng->totalArticles;
	unreadArticles=job->ng->unreadArticles;
// 	qDebug("Expiring");
	e=new UpdateDbThreadEvent(job, UpdateDbThreadEvent::UpDb_startExp);
	QApplication::postEvent(parent, e);
	if (! expire(job->ng,job->qId))
		return false;
	e=new UpdateDbThreadEvent(job, UpdateDbThreadEvent::UpDb_startUpd);
	QApplication::postEvent(parent, e);
// 	qDebug("Expired");
	Header *h;
	int tret;
	int lineProcessed=0;
	int totalLines=job->artSize;
	headerFile=new QFile(job->fName);
	//TODO: only insert in Db if artSize >= 0!
	if (totalLines != 0) {
		Db* db=job->ng->getDb();
		int hostId=job->qId;
		
		if (!headerFile->open(IO_ReadOnly)) {
			qDebug("UpdateDbThread::update: Cannot open file %s", (const char *) job->fName);
			return false;
		} 
		QTime previous, current;
		previous=QTime::currentTime();
		while ( (headerFile->readLine(line, 2000)) != -1 ) {
			h=new Header(line);
			
			lineProcessed++;
			if (h->isOk()) {
				if ((tret=dbBinHeaderPut(db, h, hostId)) != 0 && tret != DB_KEYEXIST) {
					qDebug("NntpThread::getXover(): Exit from insert: %d", tret);
					return false;
				} else job->ng->high[hostId] = h->m_num;
			} else kdDebug() << "Wrong header - maybe an incomplete line?\n";
			
			job->artSize=0;
			delete h;
			
			//DEBUG!!!
	// 		if (lineProcessed >= 50)
	// 			return false;
			//End debug!
			current=QTime::currentTime();
			if (previous.secsTo(current) > 1) {
				e=new UpdateDbThreadEvent(job, UpdateDbThreadEvent::UpDb_Working, lineProcessed, totalLines);
				QApplication::postEvent(parent, e);
				previous=QTime::currentTime();
				
			}
				
			
		}
		db->sync(0);
	} else kdDebug() << "Empty headerFile!\n";
	
	kdDebug() << "Inserted articles \'til " << job->ng->high[job->qId] << endl;
	headerFile->close();
	
	if (!headerFile->remove()) 
		qDebug("Error deleting file");
	delete headerFile;
	
	
	Dbt groupkey, groupdata;
	memset(&groupkey, 0, sizeof(groupkey));
	memset(&groupdata, 0,  sizeof(groupdata));
	job->ng->totalArticles=groupArticles;
	job->ng->unreadArticles=unreadArticles;
	char *tempng=job->ng->data();
	groupdata.set_data(tempng);
	groupdata.set_size(job->ng->getRecordSize());
	const char *tempkey=(const char *) job->ng->ngName;
	groupkey.set_data((void*)tempkey);
	groupkey.set_size(job->ng->ngName.length());
	if ((tret=job->gdb->put(NULL, &groupkey, &groupdata, 0)) != 0) {
		qDebug("Error updating newsgroup: %d", tret);
		delete tempng;
		return false;
	}
	delete tempng;
	
	job->gdb->sync(0);
	return true;
	
	
	
}

int UpdateDbThread::dbBinHeaderPut( Db * db, Header * h, int hostId ) {
    int ret;

	Dbt key, data;
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	data.set_flags(DB_DBT_MALLOC);
    BinHeader *bh;
    QRegExp rx("\\((\\d+)/(\\d+)\\)");
	
	int pos=0;
	int capPart=0;
	int capTotal=0;
	int index=-1; // = rx.search(h->m_subj, -1);
	while ( (pos=rx.search(h->m_subj, pos)) != -1 ) {
		index=pos;
		pos+=rx.matchedLength();
		capPart=rx.cap(1).toInt();
		capTotal=rx.cap(2).toInt();
	}
	
    if (index != -1) {	//probably a binheader, now add to the list, or add the corresponding part
        // check for existence
        QString ts=h->m_subj.left(index).simplifyWhiteSpace(); //this is our subject/index
		
        //build the key
        QString index=ts+h->m_from; //Now our index is: subj+from

        const char *k=(const char *) index;
        key.set_data((void*)k);
        key.set_size(index.length());
		
		ret=db->get(NULL, &key, &data, 0);
		
		
        if ((ret == 0)) {
            //key exists, add part and save binheader again
			
            bh=new BinHeader((uchar*) data.get_data());
			free(data.get_data());
			
            
//             if (bh->addPart(capPart, h, hostId)) { //Part is not a duplicate, inserted
			
			switch (bh->addPart(capPart, h, hostId)) {
				case BinHeader::Duplicate_Part:
					break;
				case BinHeader::Unread_Status:
					//Added a part that changed the status of the post...
					unreadArticles++;
				case BinHeader::New_Part:
					uchar *p = bh->data();
					memset(&data, 0, sizeof(data));
					
					data.set_size(bh->getRecordSize());
					data.set_data(p);
					if ((ret=db->put(NULL, &key, &data, 0)) != 0) {
						qDebug("Error inserting header");
					}
					// 					qDebug("Updated header inserted");
					// 				else qDebug("Error updating record");
					delete p;
					break;
				
				
					


            } //else { //part is a duplicate, do not insert? Or insert only if bin is different? Hmmm
//                 qDebug("Got a duplicate part!");
            //}
            delete bh;
			
			
            return ret;

            //debug: exit first time you make this cycle



        }
        else if (ret==DB_NOTFOUND) {
            //key doesn't exist, create header and save it
            //             				qDebug("Header does not exist, creating it");
			
            bh=new BinHeader;
			groupArticles++;
			//The new article is, of course, unread...
			unreadArticles++;
            bh->setSubj(ts);
			

            bh->setFrom(h->m_from);
            bh->setStatus(BinHeader::bh_new);



            if (h->m_date.isNull()) {
                qDebug("Date is null!!!");
                // 				qDebug("Header: %s", (const char *) h->m_subj);
                // 				qDebug("Number: %s", (const char *) h->m_num);
            } else
                bh->setDate(createDateTime(QStringList::split(" ", h->m_date, true)));

            // 			qDebug("DateTime is: %s", (const char *) bh->getDate());
            //             				qDebug("Header subj: %s", (const char *) ts);
            bh->setNumParts(capTotal);
            //             				qDebug("Parts: %d", rx.cap(2).toInt());
            bh->addPart(capPart, h, hostId);

            //             				qDebug("Part in question: %d", rx.cap(1).toInt());
            uchar *p=bh->data();
            //             				qDebug("Binheader size: %d", bh->getRecordSize());

            //             				bh->printServerPart();
			
			memset(&data, 0, sizeof(data));
            data.set_size(bh->getRecordSize());
			data.set_data(p);

            ret=db->put(NULL, &key, &data, 0);


            delete bh;
            delete p;
// 			free(data->get_data());
			
            return ret;

        
        } else {
			qDebug("Other error: %d", ret);
			return ret;
		}






    }
	return 0;








}

UpdateDbThread::UpdateDbThread( QWidget *w, QMutex *lock, QValueList< Job * > *jobs )
{
	parent=w;
	listLock=lock;
	jobList=jobs;
	
}

QString UpdateDbThread::createDateTime( QStringList dateTime) {

    int i=(dateTime[0].at(dateTime[0].length()-1) == ',') ? 1:0;
    QDate d=QDate::fromString(" " + dateTime[i+1] + " " + dateTime[i] +" " + dateTime[i+2]);
    QTime t=QTime::fromString(dateTime[i+3]);
    QDateTime dt(d, t);


    
    return dt.toString();


}
